@@ -27,7 +27,7 @@ from .utils.csp_connect import connect_to_csp
 from .utils.dbconn import sql_enquote_id
 from .utils.df_utils import CloseAfterIteration, load_dataframe_from_sql, LoadStrategy
 from .utils.file_utils import write_stream_with_read_on_close, connect_to_filesystem, adapt_fsspec_storage_options
-from .utils.format_sniffer import find_url_extension, detect_format
+from .utils.format_sniffer import find_url_extension, detect_format, FORMAT_ALIASES_WRITE_OPTIONS
 from .utils.nosql import connect_to_nosql
 from .utils.packages import install_packages
 
@@ -580,13 +580,16 @@ class DSLibrary(object):
         use_mlflow = self._mlflow_all
         # map resource name using supplied specs
         uri, kwargs = self._xlt_resource(resource_name, _mode="outputs", **kwargs)
-        file_extension = find_url_extension(uri).strip(".")
+        file_extension = find_url_extension(uri)
         if not file_extension:
             self._default_format_for_aliases(resource_name, kwargs)
         format = kwargs.pop("format", None) or file_extension
         format_options = dict(kwargs.pop("format_options", None) or {})
         if not format:
             format = format_options.pop("format", None)
+        # file extension mapping and implied options
+        format, alias_options = FORMAT_ALIASES_WRITE_OPTIONS.get(format, (format, {}))
+        format_options.update(alias_options)
         # detect SQL target
         if kwargs.get("sql_table"):
             # write data to an SQL table
@@ -669,8 +672,30 @@ class DSLibrary(object):
                         with opener(uri, 'wb', **kwargs) as f_w:
                             shutil.copyfileobj(f_r, f_w)
                 return
+            compression = format_options.pop("compression", None)
+            if compression:
+                mode += "b"
             with opener(uri, mode, **kwargs) as f_w:
-                df_method(f_w, **format_options)
+                # NOTE: pandas won't write compressed data to an open stream!
+                if compression == "zip":
+                    import zipfile
+                    with zipfile.ZipFile(f_w, mode='w', compression=zipfile.ZIP_DEFLATED) as zip_out:
+                        fn_out = uri
+                        if fn_out.endswith(".zip"):
+                            fn_out = fn_out[:-4]
+                        def xfer(f_r):
+                            zip_out.write(f_r.name, fn_out)
+                        f_wz = write_stream_with_read_on_close('wb', 'rb', xfer)
+                        df_method(f_wz, **format_options)
+                        f_wz.close()
+                elif compression:
+                    import gzip
+                    with gzip.GzipFile(fileobj=f_w, mode='w') as zipstream:
+                        df_method(zipstream, **format_options)
+                elif compression:
+                    raise Exception("only 'zip'/'gzip' compression is supported")
+                else:
+                    df_method(f_w, **format_options)
             return
         # save matplotlib figures as images
         elif hasattr(content, "savefig"):
@@ -835,7 +860,7 @@ class DSLibrary(object):
             sql_table - name of table to fully load
             nosql_collection - name of a NoSQL collection to fully load
             run_data - True to read using open_run_data()
-            dask - True to load into a dask dataframe, or an integer to load into dask if file size is >= to threshold.
+            dask - True to load into a dask dataframe, or an integer to load into dask if file size is >= to threshold (or unknown).
             sniff_only - True to do all format analysis and return a LoadStrategy, rather than a dataframe
             sniff_opener - A custom file open method for sniffing
 
@@ -960,8 +985,8 @@ class DSLibrary(object):
             if format_info.size is not None:
                 use_dask = format_info.size > use_dask
             else:
-                # size unknown - use dask if dask limit was set to a particularly low value
-                use_dask = 10000 > use_dask
+                # size unknown - use dask
+                use_dask = True
         # various formats and their read function
         if use_dask:
             use_dask = format in DASK_READ_FORMATS
