@@ -8,6 +8,10 @@ import typing
 import subprocess
 import sys
 import tempfile
+import time
+import re
+from collections import defaultdict
+
 import pandas
 
 from pexpect import split_command_line
@@ -44,6 +48,14 @@ class ModelRunner(object):
         self.data = {}
         self.parameters = parameters or {}
         self.code_paths = []
+        self._notebook_options = {}
+        self.captured_notebooks = []
+
+    def _is_local(self):
+        """
+        Detect local operation, i.e. all targets are local, there is no proxy/gateway/data service/etc.
+        """
+        return "MMLibrary" not in dslibrary._DEFAULT.__class__.__name__
 
     def _to_absolute(self, uri: str):
         """
@@ -77,7 +89,7 @@ class ModelRunner(object):
         :param kwargs:  Additional parameters to support the various data sources, following fsspec for file-like
                         sources.
         """
-        self.inputs[name] = {"uri": self._to_absolute(uri), **kwargs}
+        self.inputs[name] = {"uri": self._to_absolute(uri) if self._is_local() else uri, **kwargs}
         if data is not None:
             self.data[name] = data
         return self
@@ -96,7 +108,7 @@ class ModelRunner(object):
         :param capture: Output content from write_resource() can be captured in memory.
         :param kwargs:  See set_input()
         """
-        spec = {"uri": self._to_absolute(uri), **kwargs}
+        spec = {"uri": self._to_absolute(uri) if self._is_local() else uri, **kwargs}
         self.outputs[name] = spec
         if capture:
             spec["capture"] = True
@@ -125,6 +137,56 @@ class ModelRunner(object):
             self.code_paths.append(path)
         return self
 
+    def notebook_options(self, capture_notebook: (bool, str)=None, emit_streams: bool=None):
+        """
+        When executing notebooks you can...
+
+        :param capture_notebook: Store the complete executed notebooks in new files with timestamps.  True to generate
+                                 a filename automatically, or str to supply a specific filename.
+        :param emit_streams:     Pass through (print) the model's stdout and stderr.
+        """
+        if capture_notebook is not None:
+            self._notebook_options["capture_notebook"] = capture_notebook
+        if emit_streams is not None:
+            self._notebook_options["emit_streams"] = emit_streams
+        return self
+
+    def _capture_notebook(self, source_path: str):
+        """
+        Generate a filename for an executed notebook.
+        """
+        # caller can specify a particular filename
+        capture_opt = self._notebook_options.get("capture_notebook")
+        if isinstance(capture_opt, str) and capture_opt:
+            return capture_opt
+        # or we will generate one based on the file being executed and the current time
+        t_now = time.time()
+        timestamp = time.strftime("%Y%m%d_%H%M%S", time.gmtime(t_now))
+        millis = int(t_now * 1000) % 1000
+        timestamp += f"{millis:03d}"
+        filename = re.sub(r'\.([a-z]+)$', f".{timestamp}." + r"\1", source_path)
+        # we record generated filenames in case the user wants to look them up
+        self.captured_notebooks.append(filename)
+        return filename
+
+    def _capture_streams(self, notebook_path: str, streams):
+        """
+        Capture named output streams from a notebook.
+        """
+        with open(notebook_path, 'r') as f_r:
+            nb = json.load(f_r)
+        captured = defaultdict(list)
+        for cell in nb.get("cells", []):
+            for output in cell.get("outputs", []):
+                if output.get("output_type") != "stream":
+                    continue
+                name = output.get("name")
+                if name not in streams:
+                    continue
+                text = output.get("text")
+                captured[name].extend([text] if isinstance(text, str) else text)
+        return {name: "".join(lines) for name, lines in captured.items()}
+
     def send_metrics_to(self, uri: str=None, mlflow: bool=None, capture: bool=False, **kwargs):
         """
         Determine where metrics will be stored (for read and write).  Parameters are equivalent to set_output(), in
@@ -216,6 +278,12 @@ class ModelRunner(object):
     def run_local(self, path: str, entry_point: str=None, extra_env: dict=None):
         """
         Execute a python, R or notebook based model in a subprocess.
+
+        :param path:        Local filesystem location of code to execute.
+        :param entry_point: Named entry point, which may be mapped to a command line by metadata.
+        :param extra_env:   Additional environment variables to pass.
+
+        :returns:  Captured outputs, if any.
         """
         # if project_root is specified, 'path' is relative to it, by default
         path = path or ""
@@ -244,8 +312,15 @@ class ModelRunner(object):
         else:
             folder = self.project_root or os.path.split(path)[0] or "."
         # if we're not using metadata we have to infer how to execute the code
+        capture_notebook = None
+        del_captured_notebook = False
         if not cmd:
-            cmd = self.infer_command_from_path(path)
+            if self._notebook_options.get("capture_notebook"):
+                capture_notebook = self._capture_notebook(path)
+            elif self._notebook_options.get("emit_streams"):
+                capture_notebook = tempfile.NamedTemporaryFile(delete=False, suffix=".ipynb").name
+                del_captured_notebook = True
+            cmd = self.infer_command_from_path(path, tmp_notebook=capture_notebook)
         # convert in-memory data to files
         temp_files = []
         for resource_name, data in self.data.items():
@@ -269,15 +344,37 @@ class ModelRunner(object):
                 spec["format"] = "csv"
                 spec["format_options"] = {"sep": "\t"}
         # work out environment variables
+        # FIXME when using mmlibrary we probably still want to pass through 'folder'
         env = {
-            **self.prepare_run_env(local_path=folder),
+            **self.prepare_run_env(local_path=folder if self._is_local() else None),
             **(extra_env or {})
         }
         # call the model
         # TODO report failures in a more elegant way
-        subprocess.run(cmd, cwd=folder, env=env, check=True)
-        # capture outputs
+        if cmd[0] == "nbconvert":
+            from nbconvert.nbconvertapp import NbConvertApp
+            old_env = dict(os.environ)
+            try:
+                os.environ.update(env)
+                app = NbConvertApp()
+                app.initialize(cmd[1:])
+                app.start()
+            finally:
+                os.environ.clear()
+                os.environ.update(old_env)
+        else:
+            subprocess.run(cmd, cwd=folder, env=env, check=True)
+        # captured outputs go here (see set_output(capture))
         outputs = None
+        # dump streams from notebooks if requested
+        if capture_notebook and self._notebook_options.get("emit_streams"):
+            streams = self._capture_streams(capture_notebook, ("stdout", "stderr"))
+            if "stdout" in streams:
+                print(streams["stdout"])
+            if "stderr" in streams:
+                print(streams["stderr"], file=sys.stderr)
+            if del_captured_notebook:
+                os.remove(capture_notebook)
         for resource_name, filename in output_capture_files.items():
             if not os.path.exists(filename):
                 continue
@@ -306,7 +403,7 @@ class ModelRunner(object):
         return venv_py
 
     @staticmethod
-    def infer_command_from_path(path: str):
+    def infer_command_from_path(path: str, tmp_notebook: str=None):
         """
         When we only know the name of a source file we have to work out how to execute it.
 
@@ -325,8 +422,12 @@ class ModelRunner(object):
             # TODO detect venv for R
             cmd = ["RScript", path]
         elif f_ext == ".ipynb":
-            # FIXME apply venv_py here!
-            cmd = ["nbconvert", "--to", "notebook", "--execute", "--inplace", path]
+            cmd = ["nbconvert", "--to", "notebook", "--execute", "--log-level=ERROR"]
+            if tmp_notebook:
+                cmd += [f"--output={tmp_notebook}"]
+            else:
+                cmd.append("--inplace")
+            cmd.append(path)
         else:
             raise DSLibraryException(f"unrecognized executable extension: {f_ext} in {fn}")
         return cmd
